package server.biome;

import common.biome.Biome;
import common.biome.IBiome;
import common.block.Block;
import common.block.BlockColored;
import common.block.BlockFlower;
import common.block.BlockSand;
import common.block.BlockSapling;
import common.block.BlockTallGrass;
import common.color.DyeColor;
import common.entity.animal.EntityBat;
import common.entity.animal.EntityChicken;
import common.entity.animal.EntityCow;
import common.entity.animal.EntityMouse;
import common.entity.animal.EntityPig;
import common.entity.animal.EntityRabbit;
import common.entity.animal.EntitySheep;
import common.entity.animal.EntitySquid;
import common.entity.npc.EntityArachnoid;
import common.entity.npc.EntityHaunter;
import common.entity.npc.EntityMage;
import common.entity.npc.EntitySlime;
import common.entity.npc.EntityUndead;
import common.entity.npc.EntityZombie;
import common.init.Blocks;
import common.init.WoodType;
import common.log.Log;
import common.material.Material;
import common.rng.PerlinGen;
import common.rng.Random;
import common.rng.WeightedList;
import common.util.BlockPos;
import common.world.AWorldServer;
import common.world.State;
import common.world.World;
import server.world.WorldServer;
import server.worldgen.ChunkPrimer;
import server.worldgen.FeatureGenerator;
import server.worldgen.feature.WorldGenClay;
import server.worldgen.feature.WorldGenClayExt;
import server.worldgen.feature.WorldGenSand;
import server.worldgen.foliage.FeatureDoublePlant;
import server.worldgen.foliage.WorldGenBigMushroom;
import server.worldgen.foliage.WorldGenCactus;
import server.worldgen.foliage.WorldGenDeadBush;
import server.worldgen.foliage.WorldGenFlowers;
import server.worldgen.foliage.WorldGenMushroom;
import server.worldgen.foliage.WorldGenPumpkin;
import server.worldgen.foliage.WorldGenReed;
import server.worldgen.foliage.WorldGenTallGrass;
import server.worldgen.foliage.WorldGenWaterlily;
import server.worldgen.tree.WorldGenBaseTree;
import server.worldgen.tree.WorldGenBigTree;
import server.worldgen.tree.WorldGenBirch;
import server.worldgen.tree.WorldGenDarkOak;
import server.worldgen.tree.WorldGenJungle;
import server.worldgen.tree.WorldGenPine;
import server.worldgen.tree.WorldGenSavanna;
import server.worldgen.tree.WorldGenSwamp;
import server.worldgen.tree.WorldGenTaiga2;
import server.worldgen.tree.WorldGenTree;

public abstract class GenBiome implements IBiome {
	public static final GenBiome[] BIOMES = new GenBiome[256];
	
    public static final GenBiome none = (new BiomeNone());
    
    public static final GenBiome plains = (new BiomePlains());
    public static final GenBiome desert = (new BiomeDesert(false)).setScaling(Scaling.PLAINS_LOW);
    public static final GenBiome extremeHills = (new BiomeHills(Biome.EXTREMEHILLS, false)).setScaling(Scaling.HILLS_LARGE);
    public static final GenBiome forest = (new BiomeForest(Biome.FOREST, 0));
    public static final GenBiome taiga = (new BiomeTaiga(Biome.TAIGA, 0)).setScaling(Scaling.PLAINS_MEDIUM);
    public static final GenBiome swampland = (new BiomeSwamp()).setScaling(Scaling.SEA_POND);
    public static final GenBiome river = (new BiomeWater(Biome.RIVER, true)).setScaling(Scaling.SEA_SHALLOW);
    
    public static final GenBiome exterminated = (new BiomeExterminated());
    public static final GenBiome space = (new BiomeSpace());
    
    public static final GenBiome frozenSea = (new BiomeWater(Biome.FROZENSEA, false)).enableColdBeach().setScaling(Scaling.SEA_MEDIUM);
    public static final GenBiome frozenRiver = (new BiomeWater(Biome.FROZENRIVER, true)).enableColdBeach().setScaling(Scaling.SEA_SHALLOW);
    public static final GenBiome icePlains = (new BiomeSnow(Biome.ICEPLAINS, false)).enableColdBeach().setScaling(Scaling.PLAINS_LOW);
    public static final GenBiome iceMountains = (new BiomeSnow(Biome.ICEMOUNTAINS, false)).enableColdBeach().setScaling(Scaling.HILLS_LOW);
    public static final GenBiome mushroomPlains = (new BiomeMushroom()).setScaling(Scaling.PLAINS_VARYING);
    public static final GenBiome blackened = (new BiomeBlackened());
    public static final GenBiome beach = (new BiomeBeach(false)).setScaling(Scaling.SEA_SHORE);
    public static final GenBiome desertHills = (new BiomeDesert(true)).setScaling(Scaling.HILLS_LOW);
    public static final GenBiome forestHills = (new BiomeForest(Biome.FORESTHILLS, 0)).setScaling(Scaling.HILLS_LOW);
    public static final GenBiome taigaHills = (new BiomeTaiga(Biome.TAIGAHILLS, 0)).setScaling(Scaling.HILLS_LOW);
    public static final GenBiome extremeHillsEdge = (new BiomeHills(Biome.EXTREMEHILLSEDGE, true)).setScaling(Scaling.HILLS_MEDIUM);
    public static final GenBiome jungle = (new BiomeJungle(Biome.JUNGLE, false));
    public static final GenBiome jungleHills = (new BiomeJungle(Biome.JUNGLEHILLS, false)).setScaling(Scaling.HILLS_LOW);
    public static final GenBiome jungleEdge = (new BiomeJungle(Biome.JUNGLEEDGE, true));
    public static final GenBiome sea = (new BiomeWater(Biome.SEA, false)).setScaling(Scaling.SEA_MEDIUM);
    public static final GenBiome stoneBeach = (new BiomeStoneBeach()).setScaling(Scaling.SEA_VARYING);
    public static final GenBiome coldBeach = (new BiomeBeach(true)).setScaling(Scaling.SEA_SHORE).enableColdBeach();
    public static final GenBiome birchForest = (new BiomeForest(Biome.BIRCHFOREST, 2));
    public static final GenBiome birchForestHills = (new BiomeForest(Biome.BIRCHFORESTHILLS, 2)).setScaling(Scaling.HILLS_LOW);
    public static final GenBiome roofedForest = (new BiomeForest(Biome.ROOFEDFOREST, 3));
    public static final GenBiome coldTaiga = (new BiomeTaiga(Biome.COLDTAIGA, 0)).enableColdBeach().setScaling(Scaling.PLAINS_MEDIUM);
    public static final GenBiome coldTaigaHills = (new BiomeTaiga(Biome.COLDTAIGAHILLS, 0)).enableColdBeach().setScaling(Scaling.HILLS_LOW);
    public static final GenBiome megaTaiga = (new BiomeTaiga(Biome.MEGATAIGA, 1)).setScaling(Scaling.PLAINS_MEDIUM);
    public static final GenBiome megaTaigaHills = (new BiomeTaiga(Biome.MEGATAIGAHILLS, 1)).setScaling(Scaling.HILLS_LOW);
    public static final GenBiome extremeHillsPlus = (new BiomeHills(Biome.EXTREMEHILLSPLUS, true)).setScaling(Scaling.HILLS_LARGE);
    public static final GenBiome savanna = (new BiomeSavanna(false)).setScaling(Scaling.PLAINS_LOW);
    public static final GenBiome savannaPlateau = (new BiomeSavanna(true)).setScaling(Scaling.HILLS_PLATEAU);
    
    public static final GenBiome mesa = (new BiomeMesa(Biome.MESA, false, false));
    public static final GenBiome mesaPlateau_F = (new BiomeMesa(Biome.MESAPLATEAUF, false, true)).setScaling(Scaling.HILLS_PLATEAU);
    public static final GenBiome mesaPlateau = (new BiomeMesa(Biome.MESAPLATEAU, false, false)).setScaling(Scaling.HILLS_PLATEAU);
    
    public static final GenBiome snowLand = (new BiomeSnowLand()).enableColdBeach();
    public static final GenBiome tian = (new BiomeTian()).setScaling(Scaling.VARYING_MEDIUM);
    public static final GenBiome elvenForest = (new BiomeForest(Biome.ELVENFOREST, 4));
    public static final GenBiome upperHell = (new BiomeHell(Biome.UPPERHELL, 0));
    public static final GenBiome lowerHell = (new BiomeHell(Biome.LOWERHELL, 1));
    public static final GenBiome hellHills = (new BiomeHell(Biome.HELLHILLS, 1)).setScaling(Scaling.HILLS_LARGE);
    public static final GenBiome soulPlains = (new BiomeHell(Biome.SOULPLAINS, 1)).setScaling(Scaling.SEA_POND);
    public static final GenBiome ashLand = (new BiomeHell(Biome.ASHLAND, 2)).setScaling(Scaling.PLAINS_LOW);
    public static final GenBiome moon = (new BiomeMoon()).setScaling(Scaling.PLAINS_LOW);
    public static final GenBiome chaos = (new BiomeChaos()).setScaling(Scaling.VARYING_CHAOTIC);
    
    protected static final PerlinGen TREE_NOISE = new PerlinGen(new Random(726528729282625L), 8);
    protected static final PerlinGen GRASS_NOISE = new PerlinGen(new Random(297363826225L), 1);
    protected static final FeatureDoublePlant DOUBLE_PLANT_GEN = new FeatureDoublePlant();
    
    public final Biome base;
    
    protected final WeightedList<RngSpawn> mobs = new WeightedList<RngSpawn>();
    protected final WorldGenBaseTree worldGeneratorTrees = new WorldGenBaseTree(false);
    protected final WorldGenBigTree worldGeneratorBigTree = new WorldGenBigTree(false);
    protected final WorldGenSwamp worldGeneratorSwamp = new WorldGenSwamp();
    private final FeatureGenerator clayGen = new WorldGenClay(4);
    private final FeatureGenerator sandGen = new WorldGenSand(Blocks.sand, 7);
    private final FeatureGenerator gravelAsSandGen = new WorldGenSand(Blocks.gravel, 6);
    private final WorldGenFlowers yellowFlowerGen = new WorldGenFlowers(Blocks.flower, BlockFlower.EnumFlowerType.DANDELION);
    private final FeatureGenerator mushroomBrownGen = new WorldGenMushroom(Blocks.brown_mushroom);
    private final FeatureGenerator mushroomRedGen = new WorldGenMushroom(Blocks.red_mushroom);
    private final FeatureGenerator bigMushroomGen = new WorldGenBigMushroom();
    private final FeatureGenerator reedGen = new WorldGenReed();
    private final FeatureGenerator cactusGen = new WorldGenCactus();
    private final FeatureGenerator waterlilyGen = new WorldGenWaterlily();
    private final FeatureGenerator clayGenExt = new WorldGenClayExt(32);
    
    public State topBlock = Blocks.grass.getState();
    public State fillerBlock = Blocks.dirt.getState();
    public float depth = Scaling.VARYING_LOW.depth;
    public float scale = Scaling.VARYING_LOW.scale;
    public boolean generateLakes = true;
    public boolean generateLiquids = true;
    public boolean allowColdBeach = false;
    public boolean disallowBeach = false;

    protected int waterlilyPerChunk = 0;
    protected int treesPerChunk = 0;
    protected int flowersPerChunk = 2;
    protected int grassPerChunk = 1;
    protected int deadBushPerChunk = 0;
    protected int mushroomsPerChunk = 0;
    protected int reedsPerChunk = 0;
    protected int cactiPerChunk = 0;
    protected int sandPerChunk = 1;
    protected int sandPerChunk2 = 3;
    protected int clayPerChunk = 1;
    protected int clayExtPerChunk = 0; // 10
    protected int bigMushroomsPerChunk = 0;
    
    public static GenBiome getBiome(int id)
    {
        if (id >= 0 && id < BIOMES.length)
        {
        	return BIOMES[id];
        }
        else
        {
        	Log.JNI.warn("Biom-ID ist nicht im Bereich: " + id + ", verwende " + Biome.DEF_BIOME.id + " (" + Biome.DEF_BIOME.name + ")");
            return BIOMES[Biome.DEF_BIOME.id];
        }
    }
    
    public static void setAsProvider() {
    	IBiome.setProvider(new IBiome.BiomeProvider() {
    		public final IBiome getBiome(Biome base) {
    	    	return BIOMES[base.id];
    	    }
    	});
    }

    static {
        desert.createMutatedBiome(Biome.DESERTM);
        forest.createMutatedBiome(Biome.FLOWERFOREST);
        taiga.createMutatedBiome(Biome.TAIGAM);
        swampland.createMutatedBiome(Biome.SWAMPLANDM);
        icePlains.createMutatedBiome(Biome.ICEPLAINSSPIKES);
        jungle.createMutatedBiome(Biome.JUNGLEM);
        jungleEdge.createMutatedBiome(Biome.JUNGLEEDGEM);
        coldTaiga.createMutatedBiome(Biome.COLDTAIGAM);
        savanna.createMutatedBiome(Biome.SAVANNAM);
        savannaPlateau.createMutatedBiome(Biome.SAVANNAPLATEAUM);
        mesa.createMutatedBiome(Biome.MESABRYCE);
        mesaPlateau_F.createMutatedBiome(Biome.MESAPLATEAUFM);
        mesaPlateau.createMutatedBiome(Biome.MESAPLATEAUM);
        birchForest.createMutatedBiome(Biome.BIRCHFORESTM);
        birchForestHills.createMutatedBiome(Biome.BIRCHFORESTHILLSM);
        roofedForest.createMutatedBiome(Biome.ROOFEDFORESTM);
        megaTaiga.createMutatedBiome(Biome.MEGASPRUCETAIGA);
        extremeHills.createMutatedBiome(Biome.EXTREMEHILLSM);
        extremeHillsPlus.createMutatedBiome(Biome.EXTREMEHILLSPLUSM);
        megaTaiga.createMutatedBiome(Biome.REDWOODTAIGAHILLSM);
    }

    protected GenBiome(Biome base) {
        BIOMES[base.id] = this;
        this.base = base;
        this.addMobs(this.mobs);
    }
    
    protected void addMobs(WeightedList<RngSpawn> mobs) {
    	mobs.add(new RngSpawn(EntitySheep.class, 12, 4, 4));
        mobs.add(new RngSpawn(EntityRabbit.class, 10, 3, 10));
        mobs.add(new RngSpawn(EntityPig.class, 10, 4, 4));
        mobs.add(new RngSpawn(EntityChicken.class, 10, 4, 4));
        mobs.add(new RngSpawn(EntityCow.class, 8, 4, 4));
        mobs.add(new RngSpawn(EntityArachnoid.class, 100, 4, 4));
        mobs.add(new RngSpawn(EntityZombie.class, 100, 4, 4));
        mobs.add(new RngSpawn(EntityUndead.class, 100, 4, 4));
        mobs.add(new RngSpawn(EntityHaunter.class, 100, 4, 4));
        mobs.add(new RngSpawn(EntitySlime.class, 100, 4, 4));
//        mobs.add(new Biome.RngSpawn(EntityEnder....class, 10, 1, 4));
        mobs.add(new RngSpawn(EntityMage.class, 5, 1, 1));
        mobs.add(new RngSpawn(EntitySquid.class, 10, 4, 4));
        mobs.add(new RngSpawn(EntityBat.class, 10, 8, 8));
        mobs.add(new RngSpawn(EntityMouse.class, 10, 8, 8));
    }
    
    protected final GenBiome setScaling(Scaling scaling)
    {
        return this.setScaling(scaling.depth, scaling.scale);
    }

    protected final GenBiome setScaling(float depth, float scale)
    {
        this.depth = depth;
        this.scale = scale;
        return this;
    }

    public WorldGenTree genBigTreeChance(Random rand)
    {
        return rand.chance(10) ? this.worldGeneratorBigTree : this.worldGeneratorTrees;
    }

    public WorldGenTree genBigTreeLegacy(Random rand, BlockPos pos)
    {
        int noise = (int)((TREE_NOISE.generate((double)pos.getX() * 0.5D, (double)pos.getZ() * 0.5D) / 8D + rand.doublev() * 4D + 4D) / 3D);
        return (noise > 0 && rand.chance(noise)) || (this.base.isHighHumidity() && rand.chance(3)) ? this.worldGeneratorBigTree : 
        		this.worldGeneratorTrees;
    }
    
    public FeatureGenerator getRandomWorldGenForGrass(Random rand)
    {
        return new WorldGenTallGrass(BlockTallGrass.EnumType.GRASS);
    }

    public BlockFlower.EnumFlowerType pickRandomFlower(Random rand, BlockPos pos)
    {
        return rand.rarity(3) ? BlockFlower.EnumFlowerType.DANDELION : BlockFlower.EnumFlowerType.ROSE;
    }
    
    public State getFiller() {
    	return this.fillerBlock;
    }
    
    public State getTop() {
    	return this.topBlock;
    }
    
    protected GenBiome enableColdBeach()
    {
        this.allowColdBeach = true;
        return this;
    }
    
    protected GenBiome disableBeach()
    {
        this.disallowBeach = true;
        return this;
    }

    public WeightedList<RngSpawn> getMobs()
    {
        return this.mobs;
    }
    
    public float getMobGenChance()
    {
        return 0.1F;
    }
    
    public void decorate(WorldServer world, Random rand, BlockPos pos)
    {
        for (int i = 0; i < this.sandPerChunk2; ++i)
        {
            int j = rand.chOffset();
            int k = rand.chOffset();
            this.sandGen.generate(world, rand, world.getTopSolidOrLiquidBlock(pos.add(j, 0, k)));
        }

        for (int i1 = 0; i1 < this.clayPerChunk; ++i1)
        {
            int l1 = rand.chOffset();
            int i6 = rand.chOffset();
            this.clayGen.generate(world, rand, world.getTopSolidOrLiquidBlock(pos.add(l1, 0, i6)));
        }

        for (int j1 = 0; j1 < this.sandPerChunk; ++j1)
        {
            int i2 = rand.chOffset();
            int j6 = rand.chOffset();
            this.gravelAsSandGen.generate(world, rand, world.getTopSolidOrLiquidBlock(pos.add(i2, 0, j6)));
        }

        for (int i1 = 0; i1 < this.clayExtPerChunk; ++i1)
        {
            int l1 = rand.chOffset();
            int i6 = rand.chOffset();
            this.clayGenExt.generate(world, rand, world.getTopSolidOrLiquidBlock(pos.add(l1, 0, i6)));
        }

        int k1 = this.treesPerChunk;

        if (rand.chance(10))
        {
            ++k1;
        }

        for (int j2 = 0; j2 < k1; ++j2)
        {
            int k6 = rand.chOffset();
            int l = rand.chOffset();
            WorldGenTree treeGen = this.genBigTreeChance(rand);
            treeGen.prepare();
            BlockPos blockpos = world.getHeight(pos.add(k6, 0, l));

            if (treeGen.generate(world, rand, blockpos))
            {
                treeGen.finish(world, rand, blockpos);
            }
        }

        for (int k2 = 0; k2 < this.bigMushroomsPerChunk; ++k2)
        {
            int l6 = rand.chOffset();
            int k10 = rand.chOffset();
            this.bigMushroomGen.generate(world, rand, world.getHeight(pos.add(l6, 0, k10)));
        }

        for (int l2 = 0; l2 < this.flowersPerChunk; ++l2)
        {
            int i7 = rand.chOffset();
            int l10 = rand.chOffset();
            int j14 = world.getHeight(pos.add(i7, 0, l10)).getY() + 32;

            if (j14 > 0)
            {
                int k17 = rand.zrange(j14);
                BlockPos blockpos1 = pos.add(i7, k17, l10);
                BlockFlower.EnumFlowerType blockflower$enumflowertype = this.pickRandomFlower(rand, blockpos1);
                BlockFlower blockflower = blockflower$enumflowertype.getBlockType().getBlock();

                if (blockflower.getMaterial() != Material.air)
                {
                    this.yellowFlowerGen.setGeneratedBlock(blockflower, blockflower$enumflowertype);
                    this.yellowFlowerGen.generate(world, rand, blockpos1);
                }
            }
        }

        for (int i3 = 0; i3 < this.grassPerChunk; ++i3)
        {
            int j7 = rand.chOffset();
            int i11 = rand.chOffset();
            int k14 = world.getHeight(pos.add(j7, 0, i11)).getY() * 2;

            if (k14 > 0)
            {
                int l17 = rand.zrange(k14);
                this.getRandomWorldGenForGrass(rand).generate(world, rand, pos.add(j7, l17, i11));
            }
        }

        for (int j3 = 0; j3 < this.deadBushPerChunk; ++j3)
        {
            int k7 = rand.chOffset();
            int j11 = rand.chOffset();
            int l14 = world.getHeight(pos.add(k7, 0, j11)).getY() * 2;

            if (l14 > 0)
            {
                int i18 = rand.zrange(l14);
                (new WorldGenDeadBush()).generate(world, rand, pos.add(k7, i18, j11));
            }
        }

        for (int k3 = 0; k3 < this.waterlilyPerChunk; ++k3)
        {
            int l7 = rand.chOffset();
            int k11 = rand.chOffset();
            int i15 = world.getHeight(pos.add(l7, 0, k11)).getY() * 2;

            if (i15 > 0)
            {
                int j18 = rand.zrange(i15);
                BlockPos blockpos4;
                BlockPos blockpos7;

                for (blockpos4 = pos.add(l7, j18, k11); blockpos4.getY() > 0; blockpos4 = blockpos7)
                {
                    blockpos7 = blockpos4.down();

                    if (!world.isAirBlock(blockpos7))
                    {
                        break;
                    }
                }

                this.waterlilyGen.generate(world, rand, blockpos4);
            }
        }

        for (int l3 = 0; l3 < this.mushroomsPerChunk; ++l3)
        {
            if (rand.chance(4))
            {
                int i8 = rand.chOffset();
                int l11 = rand.chOffset();
                BlockPos blockpos2 = world.getHeight(pos.add(i8, 0, l11));
                this.mushroomBrownGen.generate(world, rand, blockpos2);
            }

            if (rand.chance(8))
            {
                int j8 = rand.chOffset();
                int i12 = rand.chOffset();
                int j15 = world.getHeight(pos.add(j8, 0, i12)).getY() * 2;

                if (j15 > 0)
                {
                    int k18 = rand.zrange(j15);
                    BlockPos blockpos5 = pos.add(j8, k18, i12);
                    this.mushroomRedGen.generate(world, rand, blockpos5);
                }
            }
        }

        if (this.mushroomsPerChunk != -1 && rand.chance(4))
        {
            int i4 = rand.chOffset();
            int k8 = rand.chOffset();
            int j12 = world.getHeight(pos.add(i4, 0, k8)).getY() * 2;

            if (j12 > 0)
            {
                int k15 = rand.zrange(j12);
                this.mushroomBrownGen.generate(world, rand, pos.add(i4, k15, k8));
            }
        }

        if (this.mushroomsPerChunk != -1 && rand.chance(8))
        {
            int j4 = rand.chOffset();
            int l8 = rand.chOffset();
            int k12 = world.getHeight(pos.add(j4, 0, l8)).getY() * 2;

            if (k12 > 0)
            {
                int l15 = rand.zrange(k12);
                this.mushroomRedGen.generate(world, rand, pos.add(j4, l15, l8));
            }
        }

        for (int k4 = 0; k4 < this.reedsPerChunk; ++k4)
        {
            int i9 = rand.chOffset();
            int l12 = rand.chOffset();
            int i16 = world.getHeight(pos.add(i9, 0, l12)).getY() * 2;

            if (i16 > 0)
            {
                int l18 = rand.zrange(i16);
                this.reedGen.generate(world, rand, pos.add(i9, l18, l12));
            }
        }

        for (int l4 = 0; l4 < 10; ++l4)
        {
            int j9 = rand.chOffset();
            int i13 = rand.chOffset();
            int j16 = world.getHeight(pos.add(j9, 0, i13)).getY() * 2;

            if (j16 > 0)
            {
                int i19 = rand.zrange(j16);
                this.reedGen.generate(world, rand, pos.add(j9, i19, i13));
            }
        }

        if (rand.chance(32))
        {
            int i5 = rand.chOffset();
            int k9 = rand.chOffset();
            int j13 = world.getHeight(pos.add(i5, 0, k9)).getY() * 2;

            if (j13 > 0)
            {
                int k16 = rand.zrange(j13);
                (new WorldGenPumpkin()).generate(world, rand, pos.add(i5, k16, k9));
            }
        }

        for (int j5 = 0; j5 < this.cactiPerChunk; ++j5)
        {
            int l9 = rand.chOffset();
            int k13 = rand.chOffset();
            int l16 = world.getHeight(pos.add(l9, 0, k13)).getY() * 2;

            if (l16 > 0)
            {
                int j19 = rand.zrange(l16);
                this.cactusGen.generate(world, rand, pos.add(l9, j19, k13));
            }
        }
    }
    
    public void genTerrainBlocks(WorldServer worldIn, Random rand, ChunkPrimer chunkPrimerIn, int x, int z, double noiseVal)
    {
        this.generateBiomeTerrain(worldIn, rand, chunkPrimerIn, x, z, noiseVal);
    }

    /**
     * Given x, z coordinates, we count down all the y positions starting at height - 1 and working our way down. When we hit a
     * non-air block, we replace it with this.topBlock (default grass, descendants may set otherwise), and then a
     * relatively shallow layer of blocks of type this.fillerBlock (default dirt). A random set of blocks below y == 5
     * (but always including y == 0) is replaced with bedrock in Chunk(...).
     *  
     * If we don't hit non-air until somewhat below sea level, we top with gravel and fill down with stone.
     *  
     * If this.fillerBlock is red sand, we replace some of that with red sandstone.
     */
    public final void generateBiomeTerrain(WorldServer worldIn, Random rand, ChunkPrimer chunkPrimerIn, int x, int z, double noiseVal)
    {
        int i = worldIn.getSeaLevel();
        State worldState = worldIn.dimension.getFiller();
        Block worldBlock = worldState.getBlock();
        State worldAlt = worldIn.dimension.getAltFiller1();
        State liquid = worldIn.getSurfaceLiquid();
        boolean freeze = liquid.getBlock().getMaterial() == Material.water;
        State iblockstate = this.topBlock;
        State iblockstate1 = this.fillerBlock;
        int j = -1;
        int k = (int)(noiseVal / 3.0D + 3.0D + rand.doublev() * 0.25D);
        int l = x & 15;
        int i1 = z & 15;
        BlockPos.MutableBlockPos blockpos$mutableblockpos = new BlockPos.MutableBlockPos();

        for (int j1 = chunkPrimerIn.height - 1; j1 >= 0; --j1)
        {
            State iblockstate2 = chunkPrimerIn.get(i1, j1, l);

            if (iblockstate2.getBlock().getMaterial() == Material.air)
            {
                j = -1;
            }
            else if (iblockstate2.getBlock() == worldBlock)
            {
                if (j == -1)
                {
                    if (k <= 0)
                    {
                        iblockstate = null;
                        iblockstate1 = worldState;
                    }
                    else if (j1 >= i - 4 && j1 <= i + 1)
                    {
                        iblockstate = this.topBlock;
                        iblockstate1 = this.fillerBlock;
                    }

                    if (j1 < i && (iblockstate == null || iblockstate.getBlock().getMaterial() == Material.air))
                    {
                    	if (freeze && World.ABSOLUTE_ZERO + worldIn.getTempOffset() + this.base.getTemperature(blockpos$mutableblockpos.set(x, j1, z)) <= 0.0F)
                        {
                            iblockstate = Blocks.ice.getState();
                        }
                        else
                        {
                            iblockstate = liquid;
                        }
                    }

                    j = k;

                    if (j1 >= i - 1)
                    {
                        chunkPrimerIn.set(i1, j1, l, iblockstate);
                    }
                    else if (j1 < i - 7 - k)
                    {
                        iblockstate = null;
                        iblockstate1 = worldState;
                        chunkPrimerIn.set(i1, j1, l, worldAlt);
                    }
                    else
                    {
                        chunkPrimerIn.set(i1, j1, l, iblockstate1);
                    }
                }
                else if (j > 0)
                {
                    --j;
                    chunkPrimerIn.set(i1, j1, l, iblockstate1);

                    if (j == 0 && iblockstate1.getBlock() == Blocks.sand)
                    {
                        j = rand.zrange(4) + Math.max(0, j1 - 63);
                        iblockstate1 = iblockstate1.getValue(BlockSand.VARIANT) == BlockSand.EnumType.RED_SAND ? Blocks.stained_hardened_clay.getState().withProperty(BlockColored.COLOR, DyeColor.ORANGE) : Blocks.sandstone.getState(); //TODO: check!
                    }
                }
            }
        }
    }
    
    public boolean generateBigMushroom(AWorldServer worldIn, BlockPos pos, State state, Random rand)
    {
        worldIn.setBlockToAir(pos);
        FeatureGenerator worldgenerator = null;

        if (state.getBlock() == Blocks.brown_mushroom)
        {
            worldgenerator = new WorldGenBigMushroom(Blocks.brown_mushroom_block);
        }
        else if (state.getBlock() == Blocks.red_mushroom)
        {
            worldgenerator = new WorldGenBigMushroom(Blocks.red_mushroom_block);
        }

        if (worldgenerator != null && worldgenerator.generate((WorldServer)worldIn, rand, pos))
        {
            return true;
        }
        else
        {
            worldIn.setState(pos, state, 3);
            return false;
        }
    }

    private boolean isTypeAt(World worldIn, BlockPos pos, WoodType type)
    {
        State iblockstate = worldIn.getState(pos);
        return iblockstate.getBlock() instanceof BlockSapling && ((BlockSapling)iblockstate.getBlock()).getWoodType() == type;
    }

    private boolean isSameSaplingTypeIn(World worldIn, BlockPos pos, int xOff, int yOff, WoodType type)
    {
        return this.isTypeAt(worldIn, pos.add(xOff, 0, yOff), type) && this.isTypeAt(worldIn, pos.add(xOff + 1, 0, yOff), type) && this.isTypeAt(worldIn, pos.add(xOff, 0, yOff + 1), type) && this.isTypeAt(worldIn, pos.add(xOff + 1, 0, yOff + 1), type);
    }
    
    public void generateTree(AWorldServer worldIn, BlockPos pos, State state, Random rand)
    {
    	WoodType type = state.getBlock() instanceof BlockSapling ? ((BlockSapling)state.getBlock()).getWoodType() : WoodType.OAK;
    	State log = type == WoodType.CHERRY ? Blocks.cherry_log.getState() : // .withProperty(BlockNewLog.VARIANT, BlockPlanks.EnumType.CHERRY) :
    		(type == WoodType.MAPLE ? Blocks.maple_log.getState() /* .withProperty(BlockNewLog.VARIANT, BlockPlanks.EnumType.MAPLE) */ : Blocks.oak_log.getState());
    	State leaves = type == WoodType.CHERRY ? Blocks.cherry_leaves.getState() : 
    		(type == WoodType.MAPLE ? Blocks.maple_leaves.getState() : Blocks.oak_leaves.getState());
        FeatureGenerator worldgenerator = (FeatureGenerator)(rand.chance(10) ? new WorldGenBigTree(true, log, leaves) : new WorldGenBaseTree(true, log, leaves));
        int i = 0;
        int j = 0;
        boolean flag = false;
//        leaves = leaves.withProperty(BlockLeaves.TYPE, worldIn.getLeavesGen());

        switch (type)
        {
            case SPRUCE:
                label114:
                for (i = 0; i >= -1; --i)
                {
                    for (j = 0; j >= -1; --j)
                    {
                        if (this.isSameSaplingTypeIn(worldIn, pos, i, j, WoodType.SPRUCE))
                        {
                            worldgenerator = new WorldGenPine(false, rand.chance());
                            flag = true;
                            break label114;
                        }
                    }
                }

                if (!flag)
                {
                    j = 0;
                    i = 0;
                    worldgenerator = new WorldGenTaiga2(true);
                }

                break;

            case BIRCH:
                worldgenerator = new WorldGenBirch(true, false);
                break;

            case TIAN:
                worldgenerator = new WorldGenBigTree(true, Blocks.tian_log.getState(), Blocks.tian_leaves.getState())
                	.setHeightLimit(6, 20);
                break;

            case JUNGLE:
                State iblockstate = Blocks.jungle_log.getState(); // .withProperty(BlockOldLog.VARIANT, BlockPlanks.EnumType.JUNGLE);
                State iblockstate1 = Blocks.jungle_leaves.getState(); // .withProperty(BlockOldLeaf.VARIANT, BlockPlanks.EnumType.JUNGLE); // .withProperty(BlockLeaves.CHECK_DECAY, Boolean.valueOf(false));
                label269:

                for (i = 0; i >= -1; --i)
                {
                    for (j = 0; j >= -1; --j)
                    {
                        if (this.isSameSaplingTypeIn(worldIn, pos, i, j, WoodType.JUNGLE))
                        {
                            worldgenerator = new WorldGenJungle(true, 10, 20, iblockstate, iblockstate1);
                            flag = true;
                            break label269;
                        }
                    }
                }

                if (!flag)
                {
                    j = 0;
                    i = 0;
                    worldgenerator = new WorldGenBaseTree(true, rand.range(4, 10), iblockstate, iblockstate1, false);
                }

                break;

            case ACACIA:
                worldgenerator = new WorldGenSavanna(true);
                break;

            case DARK_OAK:
                label390:
                for (i = 0; i >= -1; --i)
                {
                    for (j = 0; j >= -1; --j)
                    {
                        if (this.isSameSaplingTypeIn(worldIn, pos, i, j, WoodType.DARK_OAK))
                        {
                            worldgenerator = new WorldGenDarkOak(true);
                            flag = true;
                            break label390;
                        }
                    }
                }

                if (!flag)
                {
                    return;
                }

            case OAK:
            case CHERRY:
            case MAPLE:
        }

        State iblockstate2 = Blocks.air.getState();

        if (flag)
        {
            worldIn.setState(pos.add(i, 0, j), iblockstate2, 4);
            worldIn.setState(pos.add(i + 1, 0, j), iblockstate2, 4);
            worldIn.setState(pos.add(i, 0, j + 1), iblockstate2, 4);
            worldIn.setState(pos.add(i + 1, 0, j + 1), iblockstate2, 4);
        }
        else
        {
            worldIn.setState(pos, iblockstate2, 4);
        }

        if (!worldgenerator.generate((WorldServer)worldIn, rand, pos.add(i, 0, j)))
        {
            if (flag)
            {
                worldIn.setState(pos.add(i, 0, j), state, 4);
                worldIn.setState(pos.add(i + 1, 0, j), state, 4);
                worldIn.setState(pos.add(i, 0, j + 1), state, 4);
                worldIn.setState(pos.add(i + 1, 0, j + 1), state, 4);
            }
            else
            {
                worldIn.setState(pos, state, 4);
            }
        }
    }
    
    public void growGrass(AWorldServer worldIn, BlockPos pos, State state, Random rand)
    {
        BlockPos blockpos = pos.up();

        for (int i = 0; i < 128; ++i)
        {
            BlockPos blockpos1 = blockpos;
            int j = 0;

            while (true)
            {
                if (j >= i / 16)
                {
                    if (worldIn.getState(blockpos1).getBlock().getMaterial() == Material.air)
                    {
                        if (rand.chance(8))
                        {
                            BlockFlower.EnumFlowerType blockflower$enumflowertype = BIOMES[worldIn.getBiomeGenForCoords(blockpos1).id].pickRandomFlower(rand, blockpos1);
                            BlockFlower blockflower = blockflower$enumflowertype.getBlockType().getBlock();
                            State iblockstate = blockflower.getState().withProperty(blockflower.getTypeProperty(), blockflower$enumflowertype);

                            if (blockflower.canBlockStay(worldIn, blockpos1, iblockstate))
                            {
                                worldIn.setState(blockpos1, iblockstate, 3);
                            }
                        }
                        else
                        {
                            State iblockstate1 = Blocks.tallgrass.getState().withProperty(BlockTallGrass.TYPE, BlockTallGrass.EnumType.GRASS);

                            if (Blocks.tallgrass.canBlockStay(worldIn, blockpos1, iblockstate1))
                            {
                                worldIn.setState(blockpos1, iblockstate1, 3);
                            }
                        }
                    }

                    break;
                }

                blockpos1 = blockpos1.add(rand.zrange(3) - 1, (rand.zrange(3) - 1) * rand.zrange(3) / 2, rand.zrange(3) - 1);

                if (worldIn.getState(blockpos1.down()).getBlock() != Blocks.grass || worldIn.getState(blockpos1).getBlock().isNormalCube())
                {
                    break;
                }

                ++j;
            }
        }
    }
    
    protected GenBiome createMutatedBiome(Biome base)
    {
        return new BiomeMutated(base, this);
    }

    public Class <? extends GenBiome > getBiomeClass()
    {
        return this.getClass();
    }
    
    public boolean isEqualTo(GenBiome biome)
    {
        return biome == this ? true : (biome == null ? false : this.getBiomeClass() == biome.getBiomeClass());
    }

    public Temperature getTempCategory()
    {
        return this.base.temperature < -12.0f ? Temperature.COLD : (this.base.temperature < 20.0f ? Temperature.MEDIUM : Temperature.WARM);
    }
}
