package server.clipboard;

import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import common.block.Block;
import common.block.BlockDoor;
import common.block.BlockRailBase;
import common.block.ITileEntityProvider;
import common.collect.Lists;
import common.init.Blocks;
import common.inventory.IInventory;
import common.tileentity.TileEntity;
import common.util.BlockPos;
import common.util.Vec3i;
import server.world.WorldServer;

public class ClipboardPlacer {
	private static class BlockEntry {
		private final BlockPos pos;
		private final ClipboardBlock block;
		
		private BlockEntry(BlockPos pos, ClipboardBlock block) {
			this.pos = pos;
			this.block = block;
		}
	}
    
    private final WorldServer world;
    private final List<BlockEntry> stage1 = Lists.newArrayList();
    private final List<BlockEntry> stage2 = Lists.newArrayList();
    private final List<BlockEntry> stage3 = Lists.newArrayList();
    
    public ClipboardPlacer(WorldServer world) {
        this.world = world;
    }
    
    public void setBlock(BlockPos location, ClipboardBlock block) {
        if(location.getY() < 0 || location.getY() > 511)
            return;
        Block type = block.getState().getBlock();
        if (ReorderRegistry.shouldPlaceLast(type)) {
            // Place torches, etc. last
        	this.stage2.add(new BlockEntry(location, block));
        } else if (ReorderRegistry.shouldPlaceFinal(type)) {
            // Place signs, reed, etc even later
        	this.stage3.add(new BlockEntry(location, block));
        } else if (ReorderRegistry.shouldPlaceLast(this.world.getState(location).getBlock())) {
            // Destroy torches, etc. first
            this.setBlockQuirk(location, new ClipboardBlock(Blocks.air.getState()));
            this.setBlockQuirk(location, block);
        } else {
        	this.stage1.add(new BlockEntry(location, block));
        }
    }
    
    private boolean setBlockQuirk(BlockPos location, ClipboardBlock block) {
        Block existing = this.world.getState(location).getBlock();
        if (existing instanceof ITileEntityProvider) {
            TileEntity tile = this.world.getTileEntity(location);
            if ((tile instanceof IInventory)) {
                IInventory inv = (IInventory) tile;
                int size = inv.getSizeInventory();
                for (int i = 0; i < size; i++) {
                    inv.setInventorySlotContents(i, null);
                }
            }
        }
        else if (existing == Blocks.ice) {
        	this.world.setBlock(location, new ClipboardBlock(Blocks.air.getState())); // Ice turns until water so this has to be done first
        }
        return this.world.setBlock(location, block);
    }
    
    public void commit() {
//    	Iterator<Entry<BlockEntry>> iterator = Iterators.concat(this.stage1.iterator(), this.stage2.iterator());
//        while (iterator.hasNext()) {
//            Map.Entry<BlockPos, EditBlock> entry = iterator.next();
//            
//        }
        for(BlockEntry entry : this.stage1) {
        	this.setBlockQuirk(entry.pos, entry.block);
        }
        this.stage1.clear();
        for(BlockEntry entry : this.stage2) {
        	this.setBlockQuirk(entry.pos, entry.block);
        }
        this.stage2.clear();
        
        final Set<BlockPos> blocks = new HashSet<BlockPos>();
        final Map<BlockPos, ClipboardBlock> blockTypes = new HashMap<BlockPos, ClipboardBlock>();
        for(BlockEntry entry : this.stage3) {
//            final BlockPos pt = entry.getKey();
            blocks.add(entry.pos);
            blockTypes.put(entry.pos, entry.block);
        }
        this.stage3.clear();

        while (!blocks.isEmpty()) {
            BlockPos current = blocks.iterator().next();
            if (!blocks.contains(current)) {
                continue;
            }

            final Deque<BlockPos> walked = new LinkedList<BlockPos>();

            while (true) {
                walked.addFirst(current);

                assert (blockTypes.containsKey(current));

                final ClipboardBlock baseBlock = blockTypes.get(current);

                final Block type = baseBlock.getState().getBlock();

                if(type instanceof BlockDoor) {
                    if (baseBlock.getState().getValue(BlockDoor.HALF) == BlockDoor.EnumDoorHalf.LOWER) {
                        // Deal with lower door halves being attached to the floor AND the upper half
                        BlockPos upperBlock = current.add(0, 1, 0);
                        if (blocks.contains(upperBlock) && !walked.contains(upperBlock)) {
                            walked.addFirst(upperBlock);
                        }
                    }
                }
                else if(type instanceof BlockRailBase) {
                    // Here, rails are hardcoded to be attached to the block below them.
                    // They're also attached to the block they're ascending towards via BlockType.getAttachment.
                    BlockPos lowerBlock = current.add(0, -1, 0);
                    if (blocks.contains(lowerBlock) && !walked.contains(lowerBlock)) {
                        walked.addFirst(lowerBlock);
                    }
                    break;
                }

                final Vec3i attachment = ReorderRegistry.getAttachment(baseBlock.getState());
                if (attachment == null) {
                    // Block is not attached to anything => we can place it
                    break;
                }

                current = current.add(attachment);

                if (!blocks.contains(current)) {
                    // We ran outside the remaining set => assume we can place blocks on this
                    break;
                }

                if (walked.contains(current)) {
                    // Cycle detected => This will most likely go wrong, but there's nothing we can do about it.
                    break;
                }
            }

            for(BlockPos pt : walked) {
                this.setBlockQuirk(pt, blockTypes.get(pt));
                blocks.remove(pt);
            }
        }
    }
}
