package server.command;

import java.lang.reflect.Array;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import common.collect.Lists;
import common.collect.Maps;
import common.util.CharValidator;
import common.util.Vec3;
import common.world.World;
import server.command.DoubleParser.DefType;

public abstract class Command implements Executable {
	private final String name;
	private final Map<String, Parameter> parameters = Maps.newHashMap();
	private final List<Parameter> argList = Lists.newArrayList();
	
	private int parPos = 0;
	private boolean parReq = true;
	
	protected Command(String name) {
		this.name = name;
	}
	
	protected Command setParamsOptional() {
		this.parReq = false;
		return this;
	}
	
	protected Command addParameter(String name, ArgCombiner<?> combiner, ArgumentParser ... parsers) {
		Parameter param = new Parameter(name, (char)0, this.parPos++, this.parReq, Lists.newArrayList(parsers), combiner);
		this.parameters.put(name, param);
		this.argList.add(param);
		return this;
	}
	
	protected Command addParameter(String name, char shortName, ArgCombiner<?> combiner, ArgumentParser ... parsers) {
		Parameter param = new Parameter(name, shortName, -1, false, Lists.newArrayList(parsers), combiner);
		this.parameters.put(name, param);
		this.argList.add(param);
		return this;
	}
	
	protected Command addParameter(String name, ArgumentParser ... parsers) {
		return this.addParameter(name, null, parsers);
	}
	
	protected Command addParameter(String name, char shortName, ArgumentParser ... parsers) {
		return this.addParameter(name, shortName, null, parsers);
	}
	
	protected Command addParameter(ArgumentParser parser) {
		return this.addParameter(parser.getName(), parser);
	}
	
	protected Command addParameter(char shortName, ArgumentParser parser) {
		return this.addParameter(parser.getName(), shortName, parser);
	}
	
	protected Command addVector(String name, boolean defaulted, boolean centered) {
		return this.addParameter(name, new ArgCombiner<Double>() {
			public Vec3 combine(Double[] values) {
				return new Vec3(values[0], values[1], values[2]);
			}
			
			public Class<?> getTypeClass() {
				return Vec3.class;
			}
			
			public Class<Double> getInputClass() {
				return Double.class;
			}
		},
		new DoubleParser("x", defaulted ? DefType.X : null, (double)(-World.MAX_SIZE), (double)World.MAX_SIZE, centered), 
		new DoubleParser("y", defaulted ? DefType.Y : null, 0.0, (double)World.HEIGHT, false),
		new DoubleParser("z", defaulted ? DefType.Z : null, (double)(-World.MAX_SIZE), (double)World.MAX_SIZE, centered));
	}
	
	protected Command addWorld(String name, boolean defaulted) {
		return this.addParameter(new WorldParser(name, false, defaulted));
	}
	
	protected Command addDimension(String name, boolean defaulted) {
		return this.addParameter(new DimensionParser(name, defaulted));
	}
	
	protected Command addInt(String name, char shortName, int min, int max, int def) {
		return this.addParameter(shortName, new IntParser(name, false, def, min, max, def));
	}
	
	protected Command addInt(String name, int min, int max, int def) {
		return this.addParameter(new IntParser(name, false, def, min, max, def));
	}
	
	protected Command addDouble(String name, char shortName, double min, double max, double def) {
		return this.addParameter(shortName, new DoubleParser(name, def, min, max, false, def));
	}
	
	protected Command addDouble(String name, double min, double max, double def) {
		return this.addParameter(new DoubleParser(name, def, min, max, false, def));
	}
	
	protected Command addInt(String name, char shortName, int min, int max) {
		return this.addParameter(shortName, new IntParser(name, false, null, min, max));
	}
	
	protected Command addInt(String name, int min, int max) {
		return this.addParameter(new IntParser(name, false, null, min, max));
	}
	
	protected Command addDouble(String name, char shortName, double min, double max) {
		return this.addParameter(shortName, new DoubleParser(name, null, min, max, false));
	}
	
	protected Command addDouble(String name, double min, double max) {
		return this.addParameter(new DoubleParser(name, null, min, max, false));
	}
	
	protected Command addFlag(String name, char shortName) {
		return this.addParameter(name, shortName);
	}
	
//	protected ScriptExecutable addFlag(String name) {
//		return this.addParameter(name, name.charAt(0));
//	}
	
	protected <T> Command addEnum(String name, T def, Class<T> clazz, T ... values) {
		return this.addParameter(new EnumParser<T>(name, clazz, def, values));
	}
	
	protected <T> Command addEnum(String name, Class<T> clazz, T ... values) {
		return this.addEnum(name, null, clazz, values);
	}
	
	protected <T> Command addEnum(String name, T def, Class<T> clazz, Collection<T> values) {
		return this.addEnum(name, def, clazz, values.toArray((T[])Array.newInstance(clazz, values.size())));
	}
	
	protected <T> Command addEnum(String name, Class<T> clazz, Collection<T> values) {
		return this.addEnum(name, null, clazz, values);
	}
	
	protected <T> Command addEnum(String name, char shortName, T def, Class<T> clazz, T ... values) {
		return this.addParameter(shortName, new EnumParser<T>(name, clazz, def, values));
	}
	
	protected Command addTag(String name) {
		return this.addParameter(new TagParser(name, null));
	}
	
	protected Command addTag(String name, char shortName) {
		return this.addParameter(shortName, new TagParser(name, null));
	}
	
	protected Command addPlayer(String name, boolean defaulted) {
		return this.addParameter(new PlayerParser(name, defaulted));
	}
	
	protected Command addPlayerEntity(String name, boolean defaulted) {
		return this.addParameter(new PlayerEntityParser(name, defaulted));
	}
	
	protected Command addEntity(String name, boolean defaulted) {
		return this.addParameter(new EntityParser(name, defaulted, false));
	}
	
	protected Command addEntity(String name, char shortName, boolean defaulted) {
		return this.addParameter(shortName, new EntityParser(name, defaulted, false));
	}
	
	protected Command addEntityList(String name, boolean defaulted) {
		return this.addParameter(new EntityListParser(name, defaulted, false));
	}
	
	protected Command addEntityList(String name, char shortName, boolean defaulted) {
		return this.addParameter(shortName, new EntityListParser(name, defaulted, false));
	}
	
	protected Command addLivingEntity(String name, boolean defaulted) {
		return this.addParameter(new EntityParser(name, defaulted, true));
	}
	
	protected Command addLivingEntityList(String name, boolean defaulted) {
		return this.addParameter(new EntityListParser(name, defaulted, true));
	}
	
	protected Command addString(String name, boolean allowEmpty, Object ... completions) {
		return this.addParameter(new StringParser(name, null, allowEmpty, null, null, null, completions));
	}
	
	protected Command addString(String name, boolean allowEmpty, StringCompleter completer) {
		return this.addParameter(new StringParser(name, null, allowEmpty, null, null, null, completer));
	}
	
	protected Command addString(String name, String def, boolean allowEmpty, Object ... completions) {
		return this.addParameter(new StringParser(name, def, allowEmpty, null, null, null, completions));
	}
	
	protected Command addString(String name, String def, boolean allowEmpty, StringCompleter completer) {
		return this.addParameter(new StringParser(name, def, allowEmpty, null, null, null, completer));
	}
	
	protected Command addString(String name, boolean allowEmpty, Integer min, Integer max, CharValidator validator) {
		return this.addParameter(new StringParser(name, null, allowEmpty, min, max, validator));
	}

	public Map<String, Parameter> getParameters() {
		return this.parameters;
	}

	public List<Parameter> getParamList() {
		return this.argList;
	}

	public String getName() {
		return this.name;
	}
}
