package server.util;

import java.util.List;

import common.collect.Lists;
import common.util.Displayable;
import common.util.Triplet;

public abstract class Form {
	private abstract class FormElement {
		protected final String name;
		
		protected FormElement(String name) {
			this.name = name;
		}

		protected abstract Object getInputData();
		protected abstract int getInputParameter();
		protected abstract boolean acceptValue(Object obj);
	}
	
	protected class Toggle extends FormElement {
		private boolean value;
		
		protected Toggle(String name, boolean value) {
			super(name);
			this.value = value;
		}

		protected Object getInputData() {
			return this.value;
		}
		
		protected int getInputParameter() {
			return 0;
		}

		protected boolean acceptValue(Object obj) {
			if(obj instanceof Boolean)
				this.value = (Boolean)obj;
			return obj instanceof Boolean;
		}
		
		public boolean get() {
			return this.value;
		}
	}
	
	protected class Switch<T> extends FormElement {
		private final T[] values;
		private final int def;
		
		private T value;
		
		protected Switch(String name, T[] values, T value) {
			super(name);
			this.values = values;
			this.value = value;
			int def = 0;
			for(int z = 0; z < values.length; z++) {
				if(values[z] == value) {
					def = z;
					break;
				}
			}
			this.def = def;
		}

		protected Object getInputData() {
			String[] strs = new String[this.values.length];
			for(int z = 0; z < strs.length; z++) {
				strs[z] = this.values[z] instanceof Displayable ? ((Displayable)this.values[z]).getDisplay() : String.valueOf(this.values[z]);
			}
			return strs;
		}
		
		protected int getInputParameter() {
			return this.def;
		}

		protected boolean acceptValue(Object obj) {
			if(obj instanceof Integer && (Integer)obj >= 0 && (Integer)obj < this.values.length) {
				this.value = this.values[(Integer)obj];
				return true;
			}
			return false;
		}

		public T get() {
			return this.value;
		}
	}
	
	protected class Field extends FormElement {
		private final int minLength;
		private final int maxLength;
		private final boolean password;
		
		private String value;
		
		protected Field(String name, String value, int min, int max, boolean password) {
			super(name);
			this.value = value;
			this.minLength = min;
			this.maxLength = max;
			this.password = password;
		}

		protected Object getInputData() {
			return this.value;
		}
		
		protected int getInputParameter() {
			return (this.password ? 0x80000000 : 0) | (this.minLength << 16) | this.maxLength;
		}

		protected boolean acceptValue(Object obj) {
			if(obj instanceof String && ((String)obj).length() >= this.minLength && ((String)obj).length() <= this.maxLength) {
				this.value = (String)obj;
				return true;
			}
			return false;
		}

		public String get() {
			return this.value;
		}
	}
	
	private final List<FormElement> inputs = Lists.newArrayList();

	public Form() {
		this.init();
	}
	
	private <T extends FormElement> T add(T elem) {
		this.inputs.add(elem);
		return elem;
	}
	
	protected Toggle addToggle(String name, boolean def) {
		return this.add(new Toggle(name, def));
	}
	
	protected <T> Switch<T> addSwitch(String name, T def, T ... values) {
		return this.add(new Switch<T>(name, values, def));
	}
	
	protected Field addField(String name, int minLength, int maxLength, String def) {
		return this.add(new Field(name, def, minLength, maxLength, false));
	}
	
	protected Field addPassword(String name, int minLength, int maxLength, String def) {
		return this.add(new Field(name, def, minLength, maxLength, true));
	}
	
	public abstract String getTitle();
	protected abstract void init();
	protected abstract void accept();
	
	protected void cancel() {
	}
	
	public final Triplet<String, Object, Integer>[] getInputList() {
		Triplet<String, Object, Integer>[] data = new Triplet[this.inputs.size()];
		for(int z = 0; z < data.length; z++) {
			data[z] = new Triplet<String, Object, Integer>(this.inputs.get(z).name, this.inputs.get(z).getInputData(), this.inputs.get(z).getInputParameter());
		}
		return data;
	}
	
	public final void accept(Object[] data) {
		if(data == null || data.length != this.inputs.size()) {
			this.cancel();
			return;
		}
		for(int z = 0; z < data.length; z++) {
			if(!this.inputs.get(z).acceptValue(data[z])) {
				this.cancel();
				return;
			}
		}
		this.accept();
	}
}
