package server.worldgen.structure;

import java.util.Iterator;
import java.util.Map;

import common.collect.Maps;
import common.nbt.NBTBase;
import common.nbt.NBTTagCompound;
import common.rng.Random;
import common.util.BlockPos;
import common.util.ChunkPos;
import common.util.LongHashMap;
import common.world.World;
import server.world.WorldServer;
import server.world.WorldServer.WorldSavedData;
import server.worldgen.ChunkPrimer;
import server.worldgen.caves.MapGenBase;

public abstract class MapGenStructure extends MapGenBase
{
    private WorldSavedData structureData;
    protected Map<Long, StructureStart> structureMap = Maps.<Long, StructureStart>newHashMap();
    
    public abstract String getStructureName();

    /**
     * Recursively called by generate()
     */
    protected final void recursiveGenerate(WorldServer worldIn, final int chunkX, final int chunkZ, int p_180701_4_, int p_180701_5_, ChunkPrimer chunkPrimerIn)
    {
        int i = (chunkX << 4) + 8;
        int j = (chunkZ << 4) + 8;
        if(i < -World.MAX_SIZE + 2048 || j < -World.MAX_SIZE + 2048 || i >= World.MAX_SIZE - 2048 || j >= World.MAX_SIZE - 2048)
        	return;
        this.initializeStructureData(worldIn);

        if (!this.structureMap.containsKey(Long.valueOf(LongHashMap.packInt(chunkX, chunkZ))))
        {
            this.rand.intv();

            if (this.canSpawnStructureAtCoords(chunkX, chunkZ))
            {
                StructureStart structurestart = this.getStructureStart(chunkX, chunkZ);
                this.structureMap.put(Long.valueOf(LongHashMap.packInt(chunkX, chunkZ)), structurestart);
                this.setStructureStart(chunkX, chunkZ, structurestart);
            }
        }
    }

    public boolean generateStructure(WorldServer worldIn, Random randomIn, ChunkPos chunkCoord)
    {
        int i = (chunkCoord.x << 4) + 8;
        int j = (chunkCoord.z << 4) + 8;
        if(i < -World.MAX_SIZE + 2048 || j < -World.MAX_SIZE + 2048 || i >= World.MAX_SIZE - 2048 || j >= World.MAX_SIZE - 2048)
        	return false;
        this.initializeStructureData(worldIn);
        boolean flag = false;

        for (StructureStart structurestart : this.structureMap.values())
        {
            if (structurestart.isSizeableStructure() && structurestart.func_175788_a(chunkCoord) && structurestart.getBoundingBox().intersectsWith(i, j, i + 15, j + 15))
            {
                structurestart.generateStructure(worldIn, randomIn, new StructureBoundingBox(i, j, i + 15, j + 15));
                structurestart.func_175787_b(chunkCoord);
                flag = true;
                this.setStructureStart(structurestart.getChunkPosX(), structurestart.getChunkPosZ(), structurestart);
            }
        }

        return flag;
    }

    public boolean isPresent(BlockPos pos)
    {
        this.initializeStructureData(this.worldObj);
        return this.func_175797_c(pos) != null;
    }

    protected StructureStart func_175797_c(BlockPos pos)
    {
        label24:

        for (StructureStart structurestart : this.structureMap.values())
        {
            if (structurestart.isSizeableStructure() && structurestart.getBoundingBox().isVecInside(pos))
            {
                Iterator<StructureComponent> iterator = structurestart.getComponents().iterator();

                while (true)
                {
                    if (!iterator.hasNext())
                    {
                        continue label24;
                    }

                    StructureComponent structurecomponent = (StructureComponent)iterator.next();

                    if (structurecomponent.getBoundingBox().isVecInside(pos))
                    {
                        break;
                    }
                }

                return structurestart;
            }
        }

        return null;
    }

    public boolean isPositionInStructure(WorldServer worldIn, BlockPos pos)
    {
        this.initializeStructureData(worldIn);

        for (StructureStart structurestart : this.structureMap.values())
        {
            if (structurestart.isSizeableStructure() && structurestart.getBoundingBox().isVecInside(pos))
            {
                return true;
            }
        }

        return false;
    }

//    public BlockPos getClosestStrongholdPos(WorldServer worldIn, BlockPos pos)
//    {
//        this.worldObj = worldIn;
//        this.initializeStructureData(worldIn);
//        this.rand.setSeed(worldIn.getSeed());
//        long i = this.rand.longv();
//        long j = this.rand.longv();
//        long k = (long)(pos.getX() >> 4) * i;
//        long l = (long)(pos.getZ() >> 4) * j;
//        this.rand.setSeed(k ^ l ^ worldIn.getSeed());
//        this.recursiveGenerate(worldIn, pos.getX() >> 4, pos.getZ() >> 4, 0, 0, (ChunkPrimer)null);
//        double d0 = Double.MAX_VALUE;
//        BlockPos blockpos = null;
//
//        for (StructureStart structurestart : this.structureMap.values())
//        {
//            if (structurestart.isSizeableStructure())
//            {
//                StructureComponent structurecomponent = (StructureComponent)structurestart.getComponents().get(0);
//                BlockPos blockpos1 = structurecomponent.getBoundingBoxCenter();
//                double d1 = blockpos1.distanceSq(pos);
//
//                if (d1 < d0)
//                {
//                    d0 = d1;
//                    blockpos = blockpos1;
//                }
//            }
//        }
//
//        if (blockpos != null)
//        {
//            return blockpos;
//        }
//        else
//        {
//            List<BlockPos> list = this.getCoordList();
//
//            if (list != null)
//            {
//                BlockPos blockpos2 = null;
//
//                for (BlockPos blockpos3 : list)
//                {
//                    double d2 = blockpos3.distanceSq(pos);
//
//                    if (d2 < d0)
//                    {
//                        d0 = d2;
//                        blockpos2 = blockpos3;
//                    }
//                }
//
//                return blockpos2;
//            }
//            else
//            {
//                return null;
//            }
//        }
//    }

//    protected List<BlockPos> getCoordList()
//    {
//        return null;
//    }

    private void initializeStructureData(WorldServer worldIn)
    {
        if (this.structureData == null)
        {
            this.structureData = worldIn.loadItemData(this.getStructureName());

            if (this.structureData == null)
            {
                this.structureData = new WorldSavedData(this.getStructureName(), new NBTTagCompound());
                worldIn.setItemData(this.getStructureName(), this.structureData);
            }
            else
            {
                NBTTagCompound tag = this.structureData.tag;

                for (String s : tag.getKeySet())
                {
                    NBTBase nbtbase = tag.getTag(s);

                    if (nbtbase.getId() == 10)
                    {
                        NBTTagCompound nbttagcompound1 = (NBTTagCompound)nbtbase;

                        if (nbttagcompound1.hasKey("ChunkX") && nbttagcompound1.hasKey("ChunkZ"))
                        {
                            int i = nbttagcompound1.getInteger("ChunkX");
                            int j = nbttagcompound1.getInteger("ChunkZ");
                            StructureStart structurestart = MapGenStructureIO.getStructureStart(nbttagcompound1, worldIn);

                            if (structurestart != null)
                            {
                                this.structureMap.put(Long.valueOf(LongHashMap.packInt(i, j)), structurestart);
                            }
                        }
                    }
                }
            }
        }
    }

    private void setStructureStart(int chunkX, int chunkZ, StructureStart start)
    {
        this.structureData.tag.setTag("[" + chunkX + "," + chunkZ + "]", start.writeStructureComponentsToNBT(chunkX, chunkZ));
        this.structureData.dirty = true;
    }

    protected abstract boolean canSpawnStructureAtCoords(int chunkX, int chunkZ);

    protected abstract StructureStart getStructureStart(int chunkX, int chunkZ);
}
